from z3 import *
from math import sqrt, ceil

Garden_X = 5
Garden_Y = 5

SIZES  = [1,2,3,4 ]
class Plant:
    PLANT_Collections = dict([size, []] for size in SIZES)

    def __init__(self, size):
        self.size = size
        self.id = len(Plant.PLANT_Collections[size])
        Plant.PLANT_Collections[size].append(self)
        self.x = Int("S_{}_{}_X".format(size, self.id))
        self.y = Int("S_{}_{}_Y".format(size, self.id))
        self.present = Bool("S_{}_{}_ext".format(size, self.id))


    def border_inv(self):
        return  And(self.x >= 0, self.x < Garden_X, self.y >=0, self.y < Garden_Y)

    def order_inv(self):
        if self.id == 0:
            return True
        else:
            last = Plant.PLANT_Collections[self.size][self.id - 1]
            choice_order = Implies(self.present, last.present)
            placement_order= Implies(self.present, Or(self.x > last.x, self.y > last.y))
            return And(choice_order, placement_order)

    # no overlapping inv
    def location_inv(self):
        results = []
        for size in SIZES:
            if size >= self.size:
                for plant in Plant.PLANT_Collections[size]:
                    if size > self.size:
                        results.append(Implies(plant.present, Not(And(self.x ==  plant.x, self.y == plant.y))))
                    if size == self.size and plant.id > self.id:
                        results.append(Implies(plant.present, not_within_distance(self, plant)))
        return Implies(self.present, And(results))

    def __repr__(self):
        return "Plant_{}_{}".format(self.size, self.id)


    def get_all_constraints(self):
        return And(self.location_inv(), self.order_inv(), self.border_inv())


def not_within_distance(plant_1, plant_2):
    results = []
    distance = max(plant_1.size, plant_2.size)
    dx =  plant_1.x - plant_2.x
    dy = plant_1.y - plant_2.y
    abs_dx = If(dx >= 0, dx, -dx)
    abs_dy = If(dy >= 0, dy, -dy)
    abs_y = distance
    while abs_y >= 0:
        f_x = sqrt(distance*distance - abs_y*abs_y)
        abs_x =  ceil(f_x)
        if f_x < abs_x:
            results.append(Implies(abs_dy <= abs_y, abs_dx >= abs_x))
        else:
            results.append(Implies(abs_dy <= abs_y, abs_dx > abs_x))
        abs_y -= 1

    return And(results)

def get_all_plants():
    results = []
    for plants in Plant.PLANT_Collections.values():
        results.extend(plants)
    return results


def get_total_score(plants):
    return Sum([If(plant.present, plant.size, 0) for plant in plants])

def print_bored_results(model, all_plants):
    board = [['-' for _ in range(Garden_X)] for _ in range(Garden_Y)]
    for plant in all_plants:
        if model.evaluate(plant.present):
            p_x = model.evaluate(plant.x).as_long()
            p_y = model.evaluate(plant.y).as_long()
            board[p_y][p_x] = str(plant.size)
        # print("{}: {} {}".format(plant.__repr__(), p_x, p_y))

    print('\n'.join([ ' '.join([val for val in row]) for row in board]))

Config = {1:10, 2:7, 3:4, 4:3}

def garden_puzzle():
    for size, nums in Config.items():
        for _ in range(nums):
            Plant(size)

    s = SolverFor('LIA')
    all_plants = get_all_plants()
    score = get_total_score(all_plants)
    current_score = 1
    all_constraints = And(And([p.get_all_constraints() for p in all_plants]))
    s.add(all_constraints)
    s.add(score >= current_score)
    print("start solving")
    while s.check() == sat:
        model = s.model()
        print("score {}".format(model.evaluate(score).as_long()))
        print_bored_results(model, all_plants)
        current_score += 1
        s.add(score >= current_score)
    else:
        print("UNSAT")

if __name__ == "__main__":
    garden_puzzle()

